<?php

/**
 * @file
 * Definition of WoW\RemoteEntityController.
 */

/**
 * This extends the EntityAPIController class, adding required special handling
 * for remote entity objects.
 *
 * To extends this controller, an entity must have the following requirements:
 *   - a 'lastFetched' field, corresponding to the timestamp it was last fetched
 *   from the service, not necessary updated, but at least fetched.
 *   - a refresh function of the form $entity_type . '_refresh' which is responsible
 *   to refresh an existing entity. This function should honor If-Modified-Since
 *   headers if needed, and saves the entity if updated.
 *   Important note: you need to set explicitely the original field before saving
 *   the entity, doing that will prevent an infinite loop between save and load
 *   mechanics.
 *   @see wow_character_refresh()
 *   @see wow_guild_refresh()
 */
abstract class WoWRemoteEntityController extends EntityAPIController implements WoWRemoteEntityControllerInterface {

  /**
   * The wow entity refresh method.
   *
   * @var string
   */
  protected $refreshMethod;

  /**
   * The wow entity refresh threshold.
   *
   * @var string
   */
  protected $refreshThreshold;

  public function __construct($entityType) {
    parent::__construct($entityType);

    $this->refreshMethod = wow_entity_refresh_method($entityType);
    $this->refreshThreshold = wow_entity_refresh_threshold($entityType);
  }

  /**
   * Implements WoW\RemoteEntityController::refreshMethod().
   */
  public function refreshMethod($method) {
    $this->refreshMethod = $method;
  }

  /**
   * Load an entity from the database.
   *
   * If the refresh method is set to loading time, calls the corresponding hook
   * responsible for refreshing an entity.
   */
  protected function attachLoad(&$queried_entities, $revision_id = FALSE) {
    $exceptions = array();

    // Get the entity refresh method.
    if ($this->refreshMethod == WOW_REFRESH_LOAD) {

      foreach ($queried_entities as $entity) {
        try {
          // Check the lastFetched value before proceding to an update.
          if ($entity->lastFetched + $this->refreshThreshold < REQUEST_TIME) {
            // Refresh the entity by calling the service.
            module_invoke($this->entityInfo['load hook'], 'refresh', $entity);
          }
        }
        catch (WoWHttpException $e) {
          // Record every exceptions thrown during the refresh.
          $exceptions[] = $e;
        }
      }
    }

    parent::attachLoad($queried_entities, $revision_id);

    if (!empty($exceptions)) {
      // Throw either a WoWHttpMultipleException if multiple entities was
      // queried - this happen when using entity_load with multiple entities -
      // either a WoWHttpException.
      throw count($queried_entities) > 1 ? WoWHttpMultipleException($exceptions) : reset($exceptions);
    }
  }

  /**
   * Implements WoW\RemoteEntityController::fetch().
   */
  public function fetch($entity, array $query = array(), array $options = array()) {
    return $this->process($entity, wow_http_request($entity->region, $entity->remotePath(), $query, $options));
  }

  /**
   * Implements WoW\RemoteEntityController::process().
   */
  public function process($entity, $response) {
    // Updates the lastFetched timestamp with the request time.
    $entity->lastFetched = $response->getDate()->getTimestamp();

    if ($response instanceof WoWHttpResult) {

      // Merges the entity with the data from service.
      $entity->merge($response->getArray());

      // Let the controller subclasses hook the result.
      $this->processResult($entity, $response);
    }
    else {
      // Let the controller subclasses hook the status.
      $this->processStatus($entity, $response);
    }

    return $entity;
  }

  /**
   * Processes a result from the service.
   *
   * @param WoWRemoteEntity $entity
   *   The entity being working on.
   * @param WoWHttpResult $result
   *   The http result.
   *
   * @see WoWGuildController
   * @see WoWCharacterController
   */
  protected function processResult($entity, $result) {
    $entity->lastModified = $result->getArray('lastModified') / 1000;
  }

  /**
   * Processes a status from the service.
   *
   * @param WoWRemoteEntity $entity
   *   The entity being working on.
   * @param WoWHttpStatus $status
   *   The http status.
   */
  protected function processStatus($entity, $status) {
    // Updates the lastModified timestamp. This will avoid the trigger of a
    // refresh when deleting an entity for instance if the refresh method is
    // set at load time.
    db_update($this->entityInfo['base table'])
      ->condition($this->idKey, $entity->{$this->idKey})
      ->fields(array('lastFetched' => $entity->lastFetched))
      ->execute();

    switch ($status->getCode()) {
      case 304:
        // The status returned is 304 Not Modified.
        // The entity was not modified since the lastModified timestamp.
        break;

      case 404:
        // The status returned is 404 Not Found.
        // The entity was existing but can't be found anymore.
        // Deletes the local entity.
        $entity->delete();

      default:
        // The status code could not be handled here.
        throw new WoWHttpException($status);
        break;
    }
  }
}
